@@ -31,10 +31,10 @@
 // STL and Boost
 #include <functional>
 #include <boost/range/algorithm/find_if.hpp>
-#include <boost/range/algorithm/lower_bound.hpp>
+#include <boost/range/algorithm/count_if.hpp>
 
 // KDE
-#include <KConfig>
+#include <KSharedConfig>
 #include <KConfigGroup>
 
 // Local
@@ -68,6 +68,12 @@ public:
     {
         using Common::Database;
         database = Database::instance(Database::ResourcesDatabase, Database::ReadOnly);
+        s_privates << this;
+    }
+
+    ~ResultModelPrivate()
+    {
+        s_privates.removeAll(this);
     }
 
     enum Fetch {
@@ -86,18 +92,7 @@ public:
             , m_clientId(clientId)
         {
             if (!m_clientId.isEmpty()) {
-                m_configFile.reset(new KConfig(
-                        QStandardPaths::writableLocation(QStandardPaths::ConfigLocation)
-                        + QStringLiteral("/kactivitymanagerd-statsrc")));
-
-                m_config = KConfigGroup(m_configFile.get(), "ResultModel-OrderingFor-" + clientId);
-
-                if (m_config.isValid()) {
-                    m_fixedItems = m_config.readEntry("kactivitiesLinkedItemsOrder",
-                                                      QStringList());
-                }
-
-                qDebug() << "Configuration activated " << m_configFile->name();
+                m_configFile = KSharedConfig::openConfig(QStringLiteral("kactivitymanagerd-statsrc"));
             }
         }
 
@@ -110,10 +105,10 @@ public:
             return m_items.size();
         }
 
-        inline void setLinkedResultPosition(const QString &resource,
+        inline void setLinkedResultPosition(const QString &resourcePath,
                                             int position)
         {
-            if (!m_config.isValid()) {
+            if (!m_orderingConfig.isValid()) {
                 qWarning() << "We can not reorder the results, no clientId was specified";
                 return;
             }
@@ -121,23 +116,21 @@ public:
             // Preconditions:
             //  - cache is ordered properly, first on the user's desired order,
             //    then on the query specified order
-            //  - the resource that needs to be moved is already in the cache
             //  - the resource that needs to be moved is a linked resource, not
             //    one that comes from the stats (there are overly many
             //    corner-cases that need to be covered in order to support
             //    reordering of the statistics-based resources)
             //  - the new position for the resource is not outside of the cache
 
-            qDebug() << "Searching for " << resource;
-            auto resourcePosition = find(resource);
-            qDebug() << "Was resource found? " << (bool)resourcePosition;
+            auto resourcePosition = find(resourcePath);
+
             if (resourcePosition) {
-                qDebug() << "What is the status? " << resourcePosition.iterator->linkStatus();
-            }
-            if (!resourcePosition
-                || resourcePosition.iterator->linkStatus() == ResultSet::Result::NotLinked) {
-                qWarning("Trying to reposition a resource that we do not have, or is not linked");
-                return;
+                if (resourcePosition.index == position) {
+                    return;
+                }
+                if (resourcePosition.iterator->linkStatus() == ResultSet::Result::NotLinked) {
+                    return;
+                }
             }
 
             // Lets make a list of linked items - we can only reorder them,
@@ -149,26 +142,51 @@ public:
                 linkedItems << item.resource();
             }
 
-            // We can not accept the new position to be outside
-            // of the linked items area
-            if (position > linkedItems.size()) {
-                position = linkedItems.size();
+            // We have two options:
+            //  - we are planning to add an item to the desired position,
+            //    but the item is not yet in the model
+            //  - we want to move an existing item
+            if (!resourcePosition
+                    || resourcePosition.iterator->linkStatus() == ResultSet::Result::NotLinked) {
+
+                linkedItems.insert(position, resourcePath);
+
+                m_fixedOrderedItems = linkedItems;
+
+            } else {
+                // We can not accept the new position to be outside
+                // of the linked items area
+                if (position >= linkedItems.size()) {
+                    position = linkedItems.size() - 1;
+                }
+
+                Q_ASSERT(resourcePosition.index == linkedItems.indexOf(resourcePath));
+                auto oldPosition = linkedItems.indexOf(resourcePath);
+
+                const auto oldLinkedItems = linkedItems;
+
+                kamd::utils::move_one(
+                        linkedItems.begin() + oldPosition,
+                        linkedItems.begin() + position);
+
+                // When we change this, the cache is not valid anymore,
+                // destinationFor will fail and we can not use it
+                m_fixedOrderedItems = linkedItems;
+
+                // We are prepared to reorder the cache
+                d->repositionResult(resourcePosition,
+                                    d->destinationFor(*resourcePosition));
             }
 
-            auto oldPosition = linkedItems.indexOf(resource);
+            m_orderingConfig.writeEntry("kactivitiesLinkedItemsOrder", m_fixedOrderedItems);
+            m_orderingConfig.sync();
 
-            kamd::utils::slide_one(
-                    linkedItems.begin() + oldPosition,
-                    linkedItems.begin() + position);
-
-            m_fixedItems = linkedItems;
-
-            m_config.writeEntry("kactivitiesLinkedItemsOrder", m_fixedItems);
-            m_config.sync();
-
-            // We are prepared to reorder the cache
-            d->repositionResult(resourcePosition,
-                                d->destinationFor(*resourcePosition));
+            // We need to notify others to reload
+            for (const auto& other: s_privates) {
+                if (other != d && other->cache.m_clientId == m_clientId) {
+                    other->fetch(FetchReset);
+                }
+            }
         }
 
         inline void debug() const
@@ -178,6 +196,29 @@ public:
             }
         }
 
+        void loadOrderingConfig(const QString &activityTag)
+        {
+            if (!m_configFile) {
+                qDebug() << "Nothing to load - the client id is empty";
+                return;
+            }
+
+            m_orderingConfig =
+                KConfigGroup(m_configFile,
+                             QStringLiteral("ResultModel-OrderingFor-") + m_clientId + activityTag);
+
+            if (m_orderingConfig.hasKey("kactivitiesLinkedItemsOrder")) {
+                // If we have the ordering defined, use it
+                m_fixedOrderedItems = m_orderingConfig.readEntry("kactivitiesLinkedItemsOrder",
+                                                                 QStringList());
+            } else {
+                // Otherwise, copy the order from the previous activity to this one
+                m_orderingConfig.writeEntry("kactivitiesLinkedItemsOrder", m_fixedOrderedItems);
+                m_orderingConfig.sync();
+
+            }
+        }
+
     private:
         ResultModelPrivate *const d;
 
@@ -185,9 +226,9 @@ public:
         int m_countLimit;
 
         QString m_clientId;
-        std::unique_ptr<KConfig> m_configFile;
-        KConfigGroup m_config;
-        QStringList m_fixedItems;
+        KSharedConfig::Ptr m_configFile;
+        KConfigGroup m_orderingConfig;
+        QStringList m_fixedOrderedItems;
 
         friend QDebug operator<< (QDebug out, const Cache &cache)
         {
@@ -199,9 +240,9 @@ public:
         }
 
     public:
-        inline const QStringList &fixedItems() const
+        inline const QStringList &fixedOrderedItems() const
         {
-            return m_fixedItems;
+            return m_fixedOrderedItems;
         }
 
         //_ Fancy iterator, find, lowerBound
@@ -231,16 +272,6 @@ public:
             {
                 return &(*iterator);
             }
-
-            // const ResultSet::Result &operator*() const
-            // {
-            //     return *iterator;
-            // }
-            //
-            // const ResultSet::Result *operator->() const
-            // {
-            //     return &(*iterator);
-            // }
         };
 
         inline FindCacheResult find(const QString &resource)
@@ -253,29 +284,45 @@ public:
                                            == resource));
         }
 
-        template <typename What, typename Predicate>
-        inline FindCacheResult lowerBound(What &&what, Predicate &&predicate)
-        {
-            return FindCacheResult(
-                this, boost::lower_bound(m_items, std::forward<What>(what),
-                                         std::forward<Predicate>(predicate)));
-        }
-
         template <typename Predicate>
-        inline FindCacheResult lowerBound(Predicate &&predicate)
+        inline FindCacheResult lowerBoundWithSkippedResource(Predicate &&lessThanPredicate)
         {
             using namespace kamd::utils::member_matcher;
-            return FindCacheResult(
-                this, boost::lower_bound(m_items, _,
-                                         std::forward<Predicate>(predicate)));
+            const int count = boost::count_if(m_items,
+                    [&] (const ResultSet::Result &result) {
+                        return lessThanPredicate(result, _);
+                    });
+
+            return FindCacheResult(this, m_items.begin() + count);
+
+
+            // using namespace kamd::utils::member_matcher;
+            //
+            // const auto position =
+            //     std::lower_bound(m_items.begin(), m_items.end(),
+            //                      _, std::forward<Predicate>(lessThanPredicate));
+            //
+            // // We seem to have found the position for the item.
+            // // The problem is that we might have found the same position
+            // // we were previously at. Since this function is usually used
+            // // to reposition the result, we might not be in a completely
+            // // sorted collection, so the next item(s) could be less than us.
+            // // We could do this with count_if, but it would be slower
+            //
+            // if (position >= m_items.cend() - 1) {
+            //     return FindCacheResult(this, position);
+            //
+            // } else if (lessThanPredicate(_, *(position + 1))) {
+            //     return FindCacheResult(this, position);
+            //
+            // } else {
+            //     return FindCacheResult(
+            //         this, std::lower_bound(position + 1, m_items.end(),
+            //                                _, std::forward<Predicate>(lessThanPredicate)));
+            // }
         }
         //^
 
-        inline int indexOf(const FindCacheResult &result)
-        {
-            return std::distance(m_items.begin(), result.iterator);
-        }
-
         inline void insertAt(const FindCacheResult &at,
                              const ResultSet::Result &result)
         {
@@ -296,7 +343,7 @@ public:
         {
             if (m_items.size() == 0) return;
 
-            d->q->beginRemoveRows(QModelIndex(), 0, m_items.size());
+            d->q->beginRemoveRows(QModelIndex(), 0, m_items.size() - 1);
             m_items.clear();
             d->q->endRemoveRows();
         }
@@ -308,7 +355,8 @@ public:
             using namespace kamd::utils::member_matcher;
 
 #if 0
-            QDBG << "===\nOld items {";
+            QDBG << "======";
+            QDBG << "Old items {";
             for (const auto& item: m_items) {
                 QDBG << item;
             }
@@ -456,7 +504,7 @@ public:
             // Check whether we got an item representing a non-existent file,
             // if so, schedule its removal from the database
             for (const auto &item: newItems) {
-                if (item.resource().startsWith('/') && !QFile(item.resource()).exists()) {
+                if (item.resource().startsWith(QLatin1Char('/')) && !QFile(item.resource()).exists()) {
                     d->q->forgetResource(item.resource());
                 }
             }
@@ -488,45 +536,56 @@ public:
         //_ Compartor that orders the linked items by user-specified order
         typedef kamd::utils::member_matcher::placeholder placeholder;
 
-        FixedItemsLessThan(const Cache &cache,
+        enum Ordering {
+            PartialOrdering,
+            FullOrdering
+        };
+
+        FixedItemsLessThan(Ordering ordering,
+                           const Cache &cache,
                            const QString &matchResource = QString())
-            : cache(cache), matchResource(matchResource)
+            : cache(cache), matchResource(matchResource), ordering(ordering)
         {
         }
 
-        bool compare (const QString &leftResource, const QString &rightResource) const
+        bool lessThan(const QString &leftResource, const QString &rightResource) const
         {
-            const bool hasLeft  = cache.fixedItems().contains(leftResource);
-            const bool hasRight = cache.fixedItems().contains(rightResource);
+            const auto fixedOrderedItems = cache.fixedOrderedItems();
+
+            const auto indexLeft  = fixedOrderedItems.indexOf(leftResource);
+            const auto indexRight = fixedOrderedItems.indexOf(rightResource);
+
+            const bool hasLeft  = indexLeft != -1;
+            const bool hasRight = indexRight != -1;
 
             return
                 ( hasLeft && !hasRight) ? true :
                 (!hasLeft &&  hasRight) ? false :
-                ( hasLeft &&  hasRight) ? cache.fixedItems().indexOf(leftResource) <
-                                          cache.fixedItems().indexOf(rightResource) :
-                false;
+                ( hasLeft &&  hasRight) ? indexLeft < indexRight :
+                (ordering == PartialOrdering ? false : leftResource < rightResource);
         }
 
         template <typename T>
         bool operator() (const T &left, placeholder) const
         {
-            return compare(left.resource(), matchResource);
+            return lessThan(left.resource(), matchResource);
         }
 
         template <typename T>
         bool operator() (placeholder, const T &right) const
         {
-            return compare(matchResource, right.resource());
+            return lessThan(matchResource, right.resource());
         }
 
         template <typename T, typename V>
         bool operator() (const T &left, const V &right) const
         {
-            return compare(left.resource(), right.resource());
+            return lessThan(left.resource(), right.resource());
         }
 
         const Cache &cache;
         const QString matchResource;
+        Ordering ordering;
         //^
     };
 
@@ -541,14 +600,17 @@ public:
         const auto lastUpdate  = result.lastUpdate();
         const auto linkStatus  = result.linkStatus();
 
+        #define FIXED_ITEMS_LESS_THAN FixedItemsLessThan(FixedItemsLessThan::PartialOrdering, cache, resource)
         #define ORDER_BY(Field) member(&ResultSet::Result::Field) > Field
         #define ORDER_BY_FULL(Field)                                           \
             (query.selection() == Terms::AllResources ?                        \
-                cache.lowerBound(FixedItemsLessThan(cache, resource)           \
+                cache.lowerBoundWithSkippedResource(                           \
+                                 FIXED_ITEMS_LESS_THAN                         \
                                  && ORDER_BY(linkStatus)                       \
                                  && ORDER_BY(Field)                            \
                                  && ORDER_BY(resource)) :                      \
-                cache.lowerBound(FixedItemsLessThan(cache, resource)           \
+                cache.lowerBoundWithSkippedResource(                           \
+                                 FIXED_ITEMS_LESS_THAN                         \
                                  && ORDER_BY(Field)                            \
                                  && ORDER_BY(resource))                        \
             )
@@ -561,6 +623,7 @@ public:
             ;
         #undef ORDER_BY
         #undef ORDER_BY_FULL
+        #undef FIXED_ITEMS_LESS_THAN
 
         return destination;
     }
@@ -571,25 +634,34 @@ public:
         cache.removeAt(result);
         q->endRemoveRows();
 
-        fetch(cache.size(), 1);
+        if (query.selection() != Terms::LinkedResources) {
+            fetch(cache.size(), 1);
+        }
     }
 
     inline void repositionResult(const Cache::FindCacheResult &result,
                                  const Cache::FindCacheResult &destination)
     {
-        using kamd::utils::slide_one;
-
         // We already have the resource in the cache
         // So, it is the time for a reshuffle
-        const int currentIndex = result.index;
+        const int oldPosition = result.index;
+        int position = destination.index;
 
-        q->dataChanged(q->index(currentIndex), q->index(currentIndex));
+        q->dataChanged(q->index(oldPosition), q->index(oldPosition));
+
+        if (oldPosition == position) {
+            return;
+        }
+
+        if (position > oldPosition) {
+            position++;
+        }
 
         bool moving
-            = q->beginMoveRows(QModelIndex(), currentIndex, currentIndex,
-                               QModelIndex(), destination.index);
+            = q->beginMoveRows(QModelIndex(), oldPosition, oldPosition,
+                               QModelIndex(), position);
 
-        slide_one(result.iterator, destination.iterator);
+        kamd::utils::move_one(result.iterator, destination.iterator);
 
         if (moving) {
             q->endMoveRows();
@@ -664,10 +736,12 @@ public:
         hasMore = (it != results.end());
 
         // We need to sort the new items for the linked resources
-        // user-defined reordering
+        // user-defined reordering. This needs only to be a partial sort,
+        // the main sorting is done by sqlite
         if (query.selection() != Terms::UsedResources) {
-            std::stable_sort(newItems.begin(), newItems.end(),
-                             FixedItemsLessThan(cache));
+            std::stable_sort(
+                newItems.begin(), newItems.end(),
+                FixedItemsLessThan(FixedItemsLessThan::PartialOrdering, cache));
         }
 
         cache.replace(newItems, from);
@@ -680,6 +754,13 @@ public:
             // and loading all from scratch
             cache.clear();
 
+            const QString activityTag =
+                query.activities().contains(CURRENT_ACTIVITY_TAG)
+                    ? (QStringLiteral("-ForActivity-") + activities.currentActivity())
+                    : QStringLiteral("-ForAllActivities");
+
+            cache.loadOrderingConfig(activityTag);
+
             fetch(0, MAX_CHUNK_LOAD_SIZE);
 
         } else if (mode == FetchReload) {
@@ -704,8 +785,6 @@ public:
     void onResultScoreUpdated(const QString &resource, double score,
                               uint lastUpdate, uint firstUpdate)
     {
-        using boost::lower_bound;
-
         QDBG << "ResultModelPrivate::onResultScoreUpdated "
              << "result added:" << resource
              << "score:" << score
@@ -735,9 +814,7 @@ public:
             item.setLastUpdate(lastUpdate);
             item.setFirstUpdate(firstUpdate);
 
-            const auto destination = destinationFor(item);
-
-            repositionResult(result, destination);
+            repositionResult(result, destinationFor(item));
 
         } else {
             // We do not have the resource in the cache,
@@ -747,8 +824,8 @@ public:
             ResultSet::Result result;
             result.setResource(resource);
 
-            result.setTitle(" ");
-            result.setMimetype(" ");
+            result.setTitle(QStringLiteral(" "));
+            result.setMimetype(QStringLiteral(" "));
             fillTitleAndMimetype(result);
 
             result.setScore(score);
@@ -783,17 +860,8 @@ public:
 
     void onResultLinked(const QString &resource)
     {
-        const auto result = cache.find(resource);
-
-        if (!result) return;
-
-        if (query.selection() == Terms::LinkedResources) {
+        if (query.selection() != Terms::UsedResources) {
             onResultScoreUpdated(resource, 0, 0, 0);
-
-        } else if (query.selection() == Terms::AllResources) {
-            result->setLinkStatus(ResultSet::Result::Linked);
-            repositionResult(result, destinationFor(*result));
-
         }
     }
 
@@ -807,9 +875,9 @@ public:
             removeResult(result);
 
         } else if (query.selection() == Terms::AllResources) {
-            result->setLinkStatus(ResultSet::Result::NotLinked);
-            repositionResult(result, destinationFor(*result));
-
+            // When the result is unlinked, it might go away or not
+            // depending on its previous usage
+            reload();
         }
     }
 
@@ -823,19 +891,21 @@ public:
     //_ Title and mimetype functions
     void fillTitleAndMimetype(ResultSet::Result &result)
     {
+        if (!database) return;
+
         auto query = database->execQuery(
-                "SELECT "
+                QStringLiteral("SELECT "
                 "title, mimetype "
                 "FROM "
                 "ResourceInfo "
                 "WHERE "
-                "targettedResource = '" + result.resource() + "'"
+                "targettedResource = '") + result.resource() + QStringLiteral("'")
                 );
 
         // Only one item at most
         for (const auto &item: query) {
-            result.setTitle(item["title"].toString());
-            result.setMimetype(item["mimetype"].toString());
+            result.setTitle(item[QStringLiteral("title")].toString());
+            result.setMimetype(item[QStringLiteral("mimetype")].toString());
         }
     }
 
@@ -877,9 +947,12 @@ public:
 
 private:
     ResultModel *const q;
+    static QList<ResultModelPrivate*> s_privates;
 
 };
 
+QList<ResultModelPrivate*> ResultModelPrivate::s_privates;
+
 ResultModel::ResultModel(Query query, QObject *parent)
     : QAbstractListModel(parent)
     , d(new ResultModelPrivate(query, QString(), this))
@@ -902,12 +975,13 @@ ResultModel::~ResultModel()
 QHash<int, QByteArray> ResultModel::roleNames() const
 {
     return {
-        { ResourceRole    , "resource" },
-        { TitleRole       , "title" },
-        { ScoreRole       , "score" },
-        { FirstUpdateRole , "created" },
-        { LastUpdateRole  , "modified" },
-        { LinkStatusRole  , "linkStatus" }
+        { ResourceRole         , "resource" },
+        { TitleRole            , "title" },
+        { ScoreRole            , "score" },
+        { FirstUpdateRole      , "created" },
+        { LastUpdateRole       , "modified" },
+        { LinkStatusRole       , "linkStatus" },
+        { LinkedActivitiesRole , "linkedActivities" }
     };
 }
 
@@ -915,22 +989,25 @@ QVariant ResultModel::data(const QModelIndex &item, int role) const
 {
     const auto row = item.row();
 
-    if (row < 0 || row >= d->cache.size()) return QVariant();
+    if (row < 0 || row >= d->cache.size()) {
+        return QVariant();
+    }
 
     const auto &result = d->cache[row];
 
-    return role == Qt::DisplayRole ? (
-               result.title() + " " +
-               result.resource() + " - " +
-               QString::number(result.linkStatus()) + " - " +
+    return role == Qt::DisplayRole ? QString(
+               result.title() + QStringLiteral(" ") +
+               result.resource() + QStringLiteral(" - ") +
+               QString::number(result.linkStatus()) + QStringLiteral(" - ") +
                QString::number(result.score())
            )
-         : role == ResourceRole    ? result.resource()
-         : role == TitleRole       ? result.title()
-         : role == ScoreRole       ? result.score()
-         : role == FirstUpdateRole ? result.firstUpdate()
-         : role == LastUpdateRole  ? result.lastUpdate()
-         : role == LinkStatusRole  ? result.linkStatus()
+         : role == ResourceRole         ? result.resource()
+         : role == TitleRole            ? result.title()
+         : role == ScoreRole            ? result.score()
+         : role == FirstUpdateRole      ? result.firstUpdate()
+         : role == LastUpdateRole       ? result.lastUpdate()
+         : role == LinkStatusRole       ? result.linkStatus()
+         : role == LinkedActivitiesRole ? result.linkedActivities()
          : QVariant()
          ;
 }
